@@ -63,6 +63,7 @@ static Timestamp Fprofile_Time(void);
 static PyObject*    Fprofile_StartTrace(PyObject* obj, PyObject* args, PyObject* kwargs);
 static PyObject*    Fprofile_Terminate(void);
 static ThreadState* Fprofile_CreateThreadState(void);
+static void         Fprofile_ResetThreadState(void);
 static void         Fprofile_ThreadTeardown(void* key);
 
 static PyObject* Fprofile_ConfigEnableMemoryTracing(void);
@@ -74,6 +75,7 @@ static int       Fprofile_FunctionTrace(PyObject* obj, PyFrameObject* frame, int
 static PyObject* Fprofile_ThreadFunctionTrace(PyObject* obj, PyFrameObject* frame, int what, PyObject* arg);
 static PyObject* Fprofile_LoggingHook(HookedFunction* hook, PyObject* args, PyObject* kwargs);
 static PyObject* Fprofile_ImportHook(HookedFunction* hook, PyObject* args, PyObject* kwargs);
+static PyObject* Fprofile_MpexitHook(HookedFunction* hook, PyObject* args, PyObject* kwargs);
 
 #define DIM(x) (sizeof(x) / sizeof(x[0]))
 
@@ -103,6 +105,12 @@ static struct {
     // it should probably not be super large.
     char argv[64];
 
+    // Force messages to be flushed immediately.
+    // NOTE: This will add significant overhead on OSes/hardware with expensive
+    // context switches, but may be useful for debugging and is temporarily
+    // being used when we're in a forked context.
+    bool immediateFlushes;
+
     // Our PID.
     pid_t pid;
 
@@ -201,55 +209,58 @@ static Timestamp Fprofile_Time(void) {
 // attach enough information about the original function to call it.
 // NOTE: These must live forever, as Python requires them to call the
 // underlying C function.
-#define HOOK_METHOD(module, name, hook)                                     \
-    static PyObject* hooked_function_##module##_##name(PyObject* s,         \
-            PyObject* args, PyObject* kwargs);                              \
-    static HookedFunction hooked_function_info_##module##_##name = {        \
-        #module,                                                            \
-        NULL,                                                               \
-        {                                                                   \
-            #name,                                                          \
-            (PyCFunction) hooked_function_##module##_##name,                \
-            METH_VARARGS | METH_KEYWORDS,                                   \
-            NULL                                                            \
-        }                                                                   \
-    };                                                                      \
-    static PyObject* hooked_function_##module##_##name(PyObject* s,         \
-            PyObject* args, PyObject* kwargs) {                             \
-        return hook(&hooked_function_info_##module##_##name, args, kwargs); \
+
+#define HOOK_METHOD(id, module, name, hook)                    \
+    static PyObject* hooked_function_##id(PyObject* s,         \
+            PyObject* args, PyObject* kwargs);                 \
+    static HookedFunction hooked_function_info_##id = {        \
+        module,                                                \
+        NULL,                                                  \
+        {                                                      \
+            name,                                              \
+            (PyCFunction) hooked_function_##id,                \
+            METH_VARARGS | METH_KEYWORDS,                      \
+            NULL                                               \
+        }                                                      \
+    };                                                         \
+    static PyObject* hooked_function_##id(PyObject* s,         \
+            PyObject* args, PyObject* kwargs) {                \
+        return hook(&hooked_function_info_##id, args, kwargs); \
     }
 
-// Get a reference to a given HookedFunction
-#define HOOK_REFERENCE(module, name) &hooked_function_info_##module##_##name
+#define HOOK_ENTRY(id) \
+    &hooked_function_info_##id
 
 // Generate the various hooking functions we need.
-HOOK_METHOD(logging, debug, Fprofile_LoggingHook);
-HOOK_METHOD(logging, log, Fprofile_LoggingHook);
-HOOK_METHOD(logging, info, Fprofile_LoggingHook);
-HOOK_METHOD(logging, warning, Fprofile_LoggingHook);
-HOOK_METHOD(logging, error, Fprofile_LoggingHook);
-HOOK_METHOD(logging, critical, Fprofile_LoggingHook);
-HOOK_METHOD(logging, fatal, Fprofile_LoggingHook);
-HOOK_METHOD(logging, exception, Fprofile_LoggingHook);
-HOOK_METHOD(builtins, print, Fprofile_LoggingHook);
-HOOK_METHOD(builtins, __import__, Fprofile_ImportHook);
+HOOK_METHOD(0, "logging", "debug",                        Fprofile_LoggingHook)
+HOOK_METHOD(1, "logging", "log",                          Fprofile_LoggingHook)
+HOOK_METHOD(2, "logging", "info",                         Fprofile_LoggingHook)
+HOOK_METHOD(3, "logging", "warning",                      Fprofile_LoggingHook)
+HOOK_METHOD(4, "logging", "error",                        Fprofile_LoggingHook)
+HOOK_METHOD(5, "logging", "critical",                     Fprofile_LoggingHook)
+HOOK_METHOD(6, "logging", "fatal",                        Fprofile_LoggingHook)
+HOOK_METHOD(7, "logging", "exception",                    Fprofile_LoggingHook)
+HOOK_METHOD(8, "builtins", "print",                       Fprofile_LoggingHook)
+HOOK_METHOD(9, "builtins", "__import__",                  Fprofile_ImportHook)
+HOOK_METHOD(10, "multiprocessing.util", "_exit_function", Fprofile_MpexitHook)
+
+// Register the hooked functions
+static HookedFunction* hooks[] = {
+    HOOK_ENTRY(0),
+    HOOK_ENTRY(1),
+    HOOK_ENTRY(2),
+    HOOK_ENTRY(3),
+    HOOK_ENTRY(4),
+    HOOK_ENTRY(5),
+    HOOK_ENTRY(6),
+    HOOK_ENTRY(7),
+    HOOK_ENTRY(8),
+    HOOK_ENTRY(9),
+    HOOK_ENTRY(10)
+};
 
 // Register the hooks we've created.
 static void Fprofile_HookFunctions(void) {
-    // XXX: Hooks must be registered here in order to actually be used!
-    HookedFunction* hooks[] = {
-        HOOK_REFERENCE(logging, debug),
-        HOOK_REFERENCE(logging, log),
-        HOOK_REFERENCE(logging, info),
-        HOOK_REFERENCE(logging, warning),
-        HOOK_REFERENCE(logging, error),
-        HOOK_REFERENCE(logging, critical),
-        HOOK_REFERENCE(logging, fatal),
-        HOOK_REFERENCE(logging, exception),
-        HOOK_REFERENCE(builtins, print),
-        HOOK_REFERENCE(builtins, __import__),
-    };
-
     for (size_t i = 0; i < DIM(hooks); i++) {
         HookedFunction* hook = hooks[i];
 
@@ -297,6 +308,10 @@ static void* logging_malloc(void* ctx, size_t size) {
         mpack_write_u64(writer, (uintptr_t) addr);
         mpack_finish_array(writer);
         mpack_finish_array(writer);
+
+        if (moduleConfiguration.immediateFlushes) {
+            mpack_writer_flush_message(writer);
+        }
     }
 
     return addr;
@@ -325,6 +340,10 @@ static void* logging_calloc(void* ctx, size_t nelem, size_t elsize) {
         mpack_write_u64(writer, (uintptr_t) addr);
         mpack_finish_array(writer);
         mpack_finish_array(writer);
+
+        if (moduleConfiguration.immediateFlushes) {
+            mpack_writer_flush_message(writer);
+        }
     }
 
     return addr;
@@ -354,6 +373,10 @@ static void* logging_realloc(void* ctx, void* old_addr, size_t new_size) {
         mpack_write_u64(writer, (uintptr_t) new_addr);
         mpack_finish_array(writer);
         mpack_finish_array(writer);
+
+        if (moduleConfiguration.immediateFlushes) {
+            mpack_writer_flush_message(writer);
+        }
     }
 
     return new_addr;
@@ -386,6 +409,10 @@ static void logging_free(void* ctx, void* old_addr) {
         mpack_write_u64(writer, (uintptr_t) old_addr);
         mpack_finish_array(writer);
         mpack_finish_array(writer);
+
+        if (moduleConfiguration.immediateFlushes) {
+            mpack_writer_flush_message(writer);
+        }
     }
 }
 
@@ -524,6 +551,10 @@ static int Fprofile_FunctionTrace(PyObject* obj, PyFrameObject* frame, int what,
             break;
     }
 
+    if (moduleConfiguration.immediateFlushes) {
+        mpack_writer_flush_message(writer);
+    }
+
     return 0;
 }
 
@@ -563,6 +594,10 @@ static PyObject* Fprofile_LoggingHook(HookedFunction* hook, PyObject* args, PyOb
 
         mpack_write_cstr(writer, log);
         mpack_finish_array(writer);
+
+        if (moduleConfiguration.immediateFlushes) {
+            mpack_writer_flush_message(writer);
+        }
     }
 
     // Delegate to the original function now that we've logged its arguments.
@@ -594,6 +629,10 @@ static PyObject* Fprofile_ImportHook(HookedFunction* hook, PyObject* args, PyObj
 
         mpack_write_cstr(writer, import);
         mpack_finish_array(writer);
+
+        if (moduleConfiguration.immediateFlushes) {
+            mpack_writer_flush_message(writer);
+        }
     }
 
     // Delegate to the original function now that we've logged its arguments.
@@ -646,6 +685,16 @@ static void Fprofile_ThreadTeardown(void* key) {
     mpack_writer_flush_message(&state->writer);
 }
 
+static PyObject* Fprofile_MpexitHook(HookedFunction* hook, PyObject* args, PyObject* kwargs) {
+    // We're a multithreaded process and we've been notified that we're exiting.
+    // Manually run our thread's teardown, as multiprocessing won't call our
+    // normal atexit() handler.
+    ThreadState* state = pthread_getspecific(Tss_Key);
+
+    Fprofile_ThreadTeardown(state);
+    return PyObject_Call(hook->original, args, kwargs);
+}
+
 ////////////////////////////////////////////////////////////////////////////////
 // Module Initialization
 ////////////////////////////////////////////////////////////////////////////////
@@ -701,6 +750,14 @@ static PyObject* Fprofile_StartTrace(PyObject* obj, PyObject* args, PyObject* kw
 
     moduleConfiguration.pid = getpid();
 
+    // Mark that we'll need to forget some information on forks.  In
+    // particular, we shouldn't think that we have a thread that's sent any
+    // information.
+    if (pthread_atfork(NULL, NULL, Fprofile_ResetThreadState) != 0) {
+        perror("Failed to register pthread_atfork() handler");
+        exit(-1);
+    }
+
     ////////////////////////////////////////////////////////////////////////////
     // Initialization Message
     ////////////////////////////////////////////////////////////////////////////
@@ -941,3 +998,41 @@ static ThreadState* Fprofile_CreateThreadState(void) {
 
     return state;
 }
+
+// We have some existing thread state that we should free and forget about
+// before resuming logging.
+// This is useful when we've just forked and want to ensure we don't reuse an
+// existing socket.
+static void Fprofile_ResetThreadState(void) {
+    if (!moduleConfiguration.started) {
+        // We haven't actually started yet, but are for some reason being asked
+        // to fork.  This appears to be OS dependent.
+        return;
+    }
+
+    ThreadState* state = pthread_getspecific(Tss_Key);
+    assert(state != NULL);
+
+    pthread_setspecific(Tss_Key, NULL);
+    close(state->socket);
+    free(state);
+
+    // Create a new thread for this process.  We don't need to start an entire
+    // new trace like with subprocess calls since we were forked and therefore
+    // already share our moduleConfiguration.  However, we do need to record
+    // the new pid since we're in a new process.
+    moduleConfiguration.pid = getpid();
+    ThreadState* newState = Fprofile_CreateThreadState();
+
+    // Flush the state of our new thread to ensure that we'll record
+    // *something* for it in the case where we quickly exit after opening the
+    // socket.
+    // This shouldn't happen (we should catch exit cases and force a flush
+    // then), but it can't hurt to be safe.
+    //
+    // TODO: We should have a different approach that doesn't require explicit
+    // flushing.  The obvious candidate is shared-mem, but that opens us up to
+    // a whole pile of issues we'd hoped to avoid.
+    mpack_writer_flush_message(&newState->writer);
+}
+
